/****************************************************************************
**
** Copyright (C) 2016 The Qt Company Ltd.
** Contact: https://www.qt.io/licensing/
**
** This file is part of the test suite of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:GPL-EXCEPT$
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see https://www.qt.io/terms-conditions. For further
** information use the contact form at https://www.qt.io/contact-us.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3 as published by the Free Software
** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
** included in the packaging of this file. Please review the following
** information to ensure the GNU General Public License requirements will
** be met: https://www.gnu.org/licenses/gpl-3.0.html.
**
** $QT_END_LICENSE$
**
****************************************************************************/


#include <QtTest/QtTest>

#include <QtXmlPatterns/QXmlNamePool>
#include <QtXmlPatterns/QXmlName>

/*!
 \class tst_QXmlNamePool
 \internal
 \since 4.4
 \brief Tests class QXmlNamePool.

 */
class tst_QXmlNamePool : public QObject
{
    Q_OBJECT

private Q_SLOTS:
    void constCorrectness() const;
    void size() const;
    void assignmentOperator() const;
};

/*!
  Check that functions have the correct const qualification.
 */
void tst_QXmlNamePool::constCorrectness() const
{
    /* No functions are const in the name pool, so do nothing. */
}

void tst_QXmlNamePool::size() const
{
    QVERIFY2(sizeof(QXmlNamePool) == sizeof(void *), "QXmlNamePool should have a d-pointer.");
}

void tst_QXmlNamePool::assignmentOperator() const
{
    // TODO
}

QTEST_MAIN(tst_QXmlNamePool)

#include "tst_qxmlnamepool.moc"

// vim: et:ts=4:sw=4:sts=4
