/****************************************************************************
**
** Copyright (C) 2016 The Qt Company Ltd.
** Contact: https://www.qt.io/licensing/
**
** This file is part of the test suite of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:GPL-EXCEPT$
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see https://www.qt.io/terms-conditions. For further
** information use the contact form at https://www.qt.io/contact-us.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3 as published by the Free Software
** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
** included in the packaging of this file. Please review the following
** information to ensure the GNU General Public License requirements will
** be met: https://www.gnu.org/licenses/gpl-3.0.html.
**
** $QT_END_LICENSE$
**
****************************************************************************/


#include <QtCore/QCoreApplication>
#include <QtCore/QDir>
#include <QtCore/QUrl>

#include "TestFundament.h"

/*!
   This constructor is protected, such
   that inheritance is necessary.
 */
TestFundament::TestFundament()
{
}

QString TestFundament::offset()
{
#if defined(Q_OS_WIN)
    /* On Windows, the executable is placed in a release/ or debug/ folder. */
    return QLatin1String("../");
#elif defined(Q_OS_MAC)
    return QLatin1String("../../../");
#else
    return QString();
#endif
}

QString TestFundament::relativeInputFile(const QString &file)
{
    return QDir::cleanPath(offset() + file);
}

QUrl TestFundament::inputFileAsURI(const QString &file)
{
    return QUrl::fromLocalFile(inputFile(file));
}

QString TestFundament::inputFile(const QString &file)
{
    if(QDir::isAbsolutePath(file))
        return file;
    else
    {
        return QDir::cleanPath(QCoreApplication::applicationDirPath()
                               + QLatin1Char('/')
                               + offset()
                               + file);
    }
}

