/****************************************************************************
** Meta object code from reading C++ file 'qqmlxmllistmodel_p.h'
**
** Created by: The Qt Meta Object Compiler version 67 (Qt 5.15.13)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include <memory>
#include "../../../../../git/src/imports/xmllistmodel/qqmlxmllistmodel_p.h"
#include <QtCore/qbytearray.h>
#include <QtCore/qmetatype.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qqmlxmllistmodel_p.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 67
#error "This file was generated using the moc from 5.15.13. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

QT_BEGIN_MOC_NAMESPACE
QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
struct qt_meta_stringdata_QQuickXmlListModel_t {
    QByteArrayData data[38];
    char stringdata0[455];
};
#define QT_MOC_LITERAL(idx, ofs, len) \
    Q_STATIC_BYTE_ARRAY_DATA_HEADER_INITIALIZER_WITH_OFFSET(len, \
    qptrdiff(offsetof(qt_meta_stringdata_QQuickXmlListModel_t, stringdata0) + ofs \
        - idx * sizeof(QByteArrayData)) \
    )
static const qt_meta_stringdata_QQuickXmlListModel_t qt_meta_stringdata_QQuickXmlListModel = {
    {
QT_MOC_LITERAL(0, 0, 18), // "QQuickXmlListModel"
QT_MOC_LITERAL(1, 19, 15), // "DefaultProperty"
QT_MOC_LITERAL(2, 35, 5), // "roles"
QT_MOC_LITERAL(3, 41, 13), // "statusChanged"
QT_MOC_LITERAL(4, 55, 0), // ""
QT_MOC_LITERAL(5, 56, 26), // "QQuickXmlListModel::Status"
QT_MOC_LITERAL(6, 83, 15), // "progressChanged"
QT_MOC_LITERAL(7, 99, 8), // "progress"
QT_MOC_LITERAL(8, 108, 12), // "countChanged"
QT_MOC_LITERAL(9, 121, 13), // "sourceChanged"
QT_MOC_LITERAL(10, 135, 10), // "xmlChanged"
QT_MOC_LITERAL(11, 146, 12), // "queryChanged"
QT_MOC_LITERAL(12, 159, 28), // "namespaceDeclarationsChanged"
QT_MOC_LITERAL(13, 188, 6), // "reload"
QT_MOC_LITERAL(14, 195, 15), // "requestFinished"
QT_MOC_LITERAL(15, 211, 15), // "requestProgress"
QT_MOC_LITERAL(16, 227, 11), // "dataCleared"
QT_MOC_LITERAL(17, 239, 14), // "queryCompleted"
QT_MOC_LITERAL(18, 254, 20), // "QQuickXmlQueryResult"
QT_MOC_LITERAL(19, 275, 10), // "queryError"
QT_MOC_LITERAL(20, 286, 6), // "object"
QT_MOC_LITERAL(21, 293, 5), // "error"
QT_MOC_LITERAL(22, 299, 3), // "get"
QT_MOC_LITERAL(23, 303, 8), // "QJSValue"
QT_MOC_LITERAL(24, 312, 5), // "index"
QT_MOC_LITERAL(25, 318, 11), // "errorString"
QT_MOC_LITERAL(26, 330, 6), // "status"
QT_MOC_LITERAL(27, 337, 6), // "Status"
QT_MOC_LITERAL(28, 344, 6), // "source"
QT_MOC_LITERAL(29, 351, 3), // "xml"
QT_MOC_LITERAL(30, 355, 5), // "query"
QT_MOC_LITERAL(31, 361, 21), // "namespaceDeclarations"
QT_MOC_LITERAL(32, 383, 40), // "QQmlListProperty<QQuickXmlLis..."
QT_MOC_LITERAL(33, 424, 5), // "count"
QT_MOC_LITERAL(34, 430, 4), // "Null"
QT_MOC_LITERAL(35, 435, 5), // "Ready"
QT_MOC_LITERAL(36, 441, 7), // "Loading"
QT_MOC_LITERAL(37, 449, 5) // "Error"

    },
    "QQuickXmlListModel\0DefaultProperty\0"
    "roles\0statusChanged\0\0QQuickXmlListModel::Status\0"
    "progressChanged\0progress\0countChanged\0"
    "sourceChanged\0xmlChanged\0queryChanged\0"
    "namespaceDeclarationsChanged\0reload\0"
    "requestFinished\0requestProgress\0"
    "dataCleared\0queryCompleted\0"
    "QQuickXmlQueryResult\0queryError\0object\0"
    "error\0get\0QJSValue\0index\0errorString\0"
    "status\0Status\0source\0xml\0query\0"
    "namespaceDeclarations\0"
    "QQmlListProperty<QQuickXmlListModelRole>\0"
    "count\0Null\0Ready\0Loading\0Error"
};
#undef QT_MOC_LITERAL

static const uint qt_meta_data_QQuickXmlListModel[] = {

 // content:
       8,       // revision
       0,       // classname
       1,   14, // classinfo
      15,   16, // methods
       8,  122, // properties
       1,  154, // enums/sets
       0,    0, // constructors
       0,       // flags
       7,       // signalCount

 // classinfo: key, value
       1,    2,

 // signals: name, argc, parameters, tag, flags
       3,    1,   91,    4, 0x06 /* Public */,
       6,    1,   94,    4, 0x06 /* Public */,
       8,    0,   97,    4, 0x06 /* Public */,
       9,    0,   98,    4, 0x06 /* Public */,
      10,    0,   99,    4, 0x06 /* Public */,
      11,    0,  100,    4, 0x06 /* Public */,
      12,    0,  101,    4, 0x06 /* Public */,

 // slots: name, argc, parameters, tag, flags
      13,    0,  102,    4, 0x0a /* Public */,
      14,    0,  103,    4, 0x08 /* Private */,
      15,    2,  104,    4, 0x08 /* Private */,
      16,    0,  109,    4, 0x08 /* Private */,
      17,    1,  110,    4, 0x08 /* Private */,
      19,    2,  113,    4, 0x08 /* Private */,

 // methods: name, argc, parameters, tag, flags
      22,    1,  118,    4, 0x02 /* Public */,
      25,    0,  121,    4, 0x02 /* Public */,

 // signals: parameters
    QMetaType::Void, 0x80000000 | 5,    4,
    QMetaType::Void, QMetaType::QReal,    7,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,

 // slots: parameters
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, QMetaType::LongLong, QMetaType::LongLong,    4,    4,
    QMetaType::Void,
    QMetaType::Void, 0x80000000 | 18,    4,
    QMetaType::Void, QMetaType::VoidStar, QMetaType::QString,   20,   21,

 // methods: parameters
    0x80000000 | 23, QMetaType::Int,   24,
    QMetaType::QString,

 // properties: name, type, flags
      26, 0x80000000 | 27, 0x00495009,
       7, QMetaType::QReal, 0x00495001,
      28, QMetaType::QUrl, 0x00495103,
      29, QMetaType::QString, 0x00495103,
      30, QMetaType::QString, 0x00495103,
      31, QMetaType::QString, 0x00495103,
       2, 0x80000000 | 32, 0x00095009,
      33, QMetaType::Int, 0x00495001,

 // properties: notify_signal_id
       0,
       1,
       3,
       4,
       5,
       6,
       0,
       2,

 // enums: name, alias, flags, count, data
      27,   27, 0x0,    4,  159,

 // enum data: key, value
      34, uint(QQuickXmlListModel::Null),
      35, uint(QQuickXmlListModel::Ready),
      36, uint(QQuickXmlListModel::Loading),
      37, uint(QQuickXmlListModel::Error),

       0        // eod
};

void QQuickXmlListModel::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    if (_c == QMetaObject::InvokeMetaMethod) {
        auto *_t = static_cast<QQuickXmlListModel *>(_o);
        (void)_t;
        switch (_id) {
        case 0: _t->statusChanged((*reinterpret_cast< QQuickXmlListModel::Status(*)>(_a[1]))); break;
        case 1: _t->progressChanged((*reinterpret_cast< qreal(*)>(_a[1]))); break;
        case 2: _t->countChanged(); break;
        case 3: _t->sourceChanged(); break;
        case 4: _t->xmlChanged(); break;
        case 5: _t->queryChanged(); break;
        case 6: _t->namespaceDeclarationsChanged(); break;
        case 7: _t->reload(); break;
        case 8: _t->requestFinished(); break;
        case 9: _t->requestProgress((*reinterpret_cast< qint64(*)>(_a[1])),(*reinterpret_cast< qint64(*)>(_a[2]))); break;
        case 10: _t->dataCleared(); break;
        case 11: _t->queryCompleted((*reinterpret_cast< const QQuickXmlQueryResult(*)>(_a[1]))); break;
        case 12: _t->queryError((*reinterpret_cast< void*(*)>(_a[1])),(*reinterpret_cast< const QString(*)>(_a[2]))); break;
        case 13: { QJSValue _r = _t->get((*reinterpret_cast< int(*)>(_a[1])));
            if (_a[0]) *reinterpret_cast< QJSValue*>(_a[0]) = std::move(_r); }  break;
        case 14: { QString _r = _t->errorString();
            if (_a[0]) *reinterpret_cast< QString*>(_a[0]) = std::move(_r); }  break;
        default: ;
        }
    } else if (_c == QMetaObject::IndexOfMethod) {
        int *result = reinterpret_cast<int *>(_a[0]);
        {
            using _t = void (QQuickXmlListModel::*)(QQuickXmlListModel::Status );
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&QQuickXmlListModel::statusChanged)) {
                *result = 0;
                return;
            }
        }
        {
            using _t = void (QQuickXmlListModel::*)(qreal );
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&QQuickXmlListModel::progressChanged)) {
                *result = 1;
                return;
            }
        }
        {
            using _t = void (QQuickXmlListModel::*)();
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&QQuickXmlListModel::countChanged)) {
                *result = 2;
                return;
            }
        }
        {
            using _t = void (QQuickXmlListModel::*)();
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&QQuickXmlListModel::sourceChanged)) {
                *result = 3;
                return;
            }
        }
        {
            using _t = void (QQuickXmlListModel::*)();
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&QQuickXmlListModel::xmlChanged)) {
                *result = 4;
                return;
            }
        }
        {
            using _t = void (QQuickXmlListModel::*)();
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&QQuickXmlListModel::queryChanged)) {
                *result = 5;
                return;
            }
        }
        {
            using _t = void (QQuickXmlListModel::*)();
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&QQuickXmlListModel::namespaceDeclarationsChanged)) {
                *result = 6;
                return;
            }
        }
    } else if (_c == QMetaObject::RegisterPropertyMetaType) {
        switch (_id) {
        default: *reinterpret_cast<int*>(_a[0]) = -1; break;
        case 6:
            *reinterpret_cast<int*>(_a[0]) = qRegisterMetaType< QQmlListProperty<QQuickXmlListModelRole> >(); break;
        }
    }

#ifndef QT_NO_PROPERTIES
    else if (_c == QMetaObject::ReadProperty) {
        auto *_t = static_cast<QQuickXmlListModel *>(_o);
        (void)_t;
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast< Status*>(_v) = _t->status(); break;
        case 1: *reinterpret_cast< qreal*>(_v) = _t->progress(); break;
        case 2: *reinterpret_cast< QUrl*>(_v) = _t->source(); break;
        case 3: *reinterpret_cast< QString*>(_v) = _t->xml(); break;
        case 4: *reinterpret_cast< QString*>(_v) = _t->query(); break;
        case 5: *reinterpret_cast< QString*>(_v) = _t->namespaceDeclarations(); break;
        case 6: *reinterpret_cast< QQmlListProperty<QQuickXmlListModelRole>*>(_v) = _t->roleObjects(); break;
        case 7: *reinterpret_cast< int*>(_v) = _t->count(); break;
        default: break;
        }
    } else if (_c == QMetaObject::WriteProperty) {
        auto *_t = static_cast<QQuickXmlListModel *>(_o);
        (void)_t;
        void *_v = _a[0];
        switch (_id) {
        case 2: _t->setSource(*reinterpret_cast< QUrl*>(_v)); break;
        case 3: _t->setXml(*reinterpret_cast< QString*>(_v)); break;
        case 4: _t->setQuery(*reinterpret_cast< QString*>(_v)); break;
        case 5: _t->setNamespaceDeclarations(*reinterpret_cast< QString*>(_v)); break;
        default: break;
        }
    } else if (_c == QMetaObject::ResetProperty) {
    }
#endif // QT_NO_PROPERTIES
}

QT_INIT_METAOBJECT const QMetaObject QQuickXmlListModel::staticMetaObject = { {
    QMetaObject::SuperData::link<QAbstractListModel::staticMetaObject>(),
    qt_meta_stringdata_QQuickXmlListModel.data,
    qt_meta_data_QQuickXmlListModel,
    qt_static_metacall,
    nullptr,
    nullptr
} };


const QMetaObject *QQuickXmlListModel::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QQuickXmlListModel::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_meta_stringdata_QQuickXmlListModel.stringdata0))
        return static_cast<void*>(this);
    if (!strcmp(_clname, "QQmlParserStatus"))
        return static_cast< QQmlParserStatus*>(this);
    if (!strcmp(_clname, "org.qt-project.Qt.QQmlParserStatus"))
        return static_cast< QQmlParserStatus*>(this);
    return QAbstractListModel::qt_metacast(_clname);
}

int QQuickXmlListModel::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QAbstractListModel::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 15)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 15;
    } else if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 15)
            *reinterpret_cast<int*>(_a[0]) = -1;
        _id -= 15;
    }
#ifndef QT_NO_PROPERTIES
    else if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 8;
    } else if (_c == QMetaObject::QueryPropertyDesignable) {
        _id -= 8;
    } else if (_c == QMetaObject::QueryPropertyScriptable) {
        _id -= 8;
    } else if (_c == QMetaObject::QueryPropertyStored) {
        _id -= 8;
    } else if (_c == QMetaObject::QueryPropertyEditable) {
        _id -= 8;
    } else if (_c == QMetaObject::QueryPropertyUser) {
        _id -= 8;
    }
#endif // QT_NO_PROPERTIES
    return _id;
}

// SIGNAL 0
void QQuickXmlListModel::statusChanged(QQuickXmlListModel::Status _t1)
{
    void *_a[] = { nullptr, const_cast<void*>(reinterpret_cast<const void*>(std::addressof(_t1))) };
    QMetaObject::activate(this, &staticMetaObject, 0, _a);
}

// SIGNAL 1
void QQuickXmlListModel::progressChanged(qreal _t1)
{
    void *_a[] = { nullptr, const_cast<void*>(reinterpret_cast<const void*>(std::addressof(_t1))) };
    QMetaObject::activate(this, &staticMetaObject, 1, _a);
}

// SIGNAL 2
void QQuickXmlListModel::countChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void QQuickXmlListModel::sourceChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void QQuickXmlListModel::xmlChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}

// SIGNAL 5
void QQuickXmlListModel::queryChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}

// SIGNAL 6
void QQuickXmlListModel::namespaceDeclarationsChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 6, nullptr);
}
struct qt_meta_stringdata_QQuickXmlListModelRole_t {
    QByteArrayData data[8];
    char stringdata0[79];
};
#define QT_MOC_LITERAL(idx, ofs, len) \
    Q_STATIC_BYTE_ARRAY_DATA_HEADER_INITIALIZER_WITH_OFFSET(len, \
    qptrdiff(offsetof(qt_meta_stringdata_QQuickXmlListModelRole_t, stringdata0) + ofs \
        - idx * sizeof(QByteArrayData)) \
    )
static const qt_meta_stringdata_QQuickXmlListModelRole_t qt_meta_stringdata_QQuickXmlListModelRole = {
    {
QT_MOC_LITERAL(0, 0, 22), // "QQuickXmlListModelRole"
QT_MOC_LITERAL(1, 23, 11), // "nameChanged"
QT_MOC_LITERAL(2, 35, 0), // ""
QT_MOC_LITERAL(3, 36, 12), // "queryChanged"
QT_MOC_LITERAL(4, 49, 12), // "isKeyChanged"
QT_MOC_LITERAL(5, 62, 4), // "name"
QT_MOC_LITERAL(6, 67, 5), // "query"
QT_MOC_LITERAL(7, 73, 5) // "isKey"

    },
    "QQuickXmlListModelRole\0nameChanged\0\0"
    "queryChanged\0isKeyChanged\0name\0query\0"
    "isKey"
};
#undef QT_MOC_LITERAL

static const uint qt_meta_data_QQuickXmlListModelRole[] = {

 // content:
       8,       // revision
       0,       // classname
       0,    0, // classinfo
       3,   14, // methods
       3,   32, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       3,       // signalCount

 // signals: name, argc, parameters, tag, flags
       1,    0,   29,    2, 0x06 /* Public */,
       3,    0,   30,    2, 0x06 /* Public */,
       4,    0,   31,    2, 0x06 /* Public */,

 // signals: parameters
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,

 // properties: name, type, flags
       5, QMetaType::QString, 0x00495103,
       6, QMetaType::QString, 0x00495103,
       7, QMetaType::Bool, 0x00495103,

 // properties: notify_signal_id
       0,
       1,
       2,

       0        // eod
};

void QQuickXmlListModelRole::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    if (_c == QMetaObject::InvokeMetaMethod) {
        auto *_t = static_cast<QQuickXmlListModelRole *>(_o);
        (void)_t;
        switch (_id) {
        case 0: _t->nameChanged(); break;
        case 1: _t->queryChanged(); break;
        case 2: _t->isKeyChanged(); break;
        default: ;
        }
    } else if (_c == QMetaObject::IndexOfMethod) {
        int *result = reinterpret_cast<int *>(_a[0]);
        {
            using _t = void (QQuickXmlListModelRole::*)();
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&QQuickXmlListModelRole::nameChanged)) {
                *result = 0;
                return;
            }
        }
        {
            using _t = void (QQuickXmlListModelRole::*)();
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&QQuickXmlListModelRole::queryChanged)) {
                *result = 1;
                return;
            }
        }
        {
            using _t = void (QQuickXmlListModelRole::*)();
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&QQuickXmlListModelRole::isKeyChanged)) {
                *result = 2;
                return;
            }
        }
    }
#ifndef QT_NO_PROPERTIES
    else if (_c == QMetaObject::ReadProperty) {
        auto *_t = static_cast<QQuickXmlListModelRole *>(_o);
        (void)_t;
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast< QString*>(_v) = _t->name(); break;
        case 1: *reinterpret_cast< QString*>(_v) = _t->query(); break;
        case 2: *reinterpret_cast< bool*>(_v) = _t->isKey(); break;
        default: break;
        }
    } else if (_c == QMetaObject::WriteProperty) {
        auto *_t = static_cast<QQuickXmlListModelRole *>(_o);
        (void)_t;
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setName(*reinterpret_cast< QString*>(_v)); break;
        case 1: _t->setQuery(*reinterpret_cast< QString*>(_v)); break;
        case 2: _t->setIsKey(*reinterpret_cast< bool*>(_v)); break;
        default: break;
        }
    } else if (_c == QMetaObject::ResetProperty) {
    }
#endif // QT_NO_PROPERTIES
    (void)_a;
}

QT_INIT_METAOBJECT const QMetaObject QQuickXmlListModelRole::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_meta_stringdata_QQuickXmlListModelRole.data,
    qt_meta_data_QQuickXmlListModelRole,
    qt_static_metacall,
    nullptr,
    nullptr
} };


const QMetaObject *QQuickXmlListModelRole::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QQuickXmlListModelRole::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_meta_stringdata_QQuickXmlListModelRole.stringdata0))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int QQuickXmlListModelRole::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 3)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 3;
    } else if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 3)
            *reinterpret_cast<int*>(_a[0]) = -1;
        _id -= 3;
    }
#ifndef QT_NO_PROPERTIES
    else if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 3;
    } else if (_c == QMetaObject::QueryPropertyDesignable) {
        _id -= 3;
    } else if (_c == QMetaObject::QueryPropertyScriptable) {
        _id -= 3;
    } else if (_c == QMetaObject::QueryPropertyStored) {
        _id -= 3;
    } else if (_c == QMetaObject::QueryPropertyEditable) {
        _id -= 3;
    } else if (_c == QMetaObject::QueryPropertyUser) {
        _id -= 3;
    }
#endif // QT_NO_PROPERTIES
    return _id;
}

// SIGNAL 0
void QQuickXmlListModelRole::nameChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void QQuickXmlListModelRole::queryChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void QQuickXmlListModelRole::isKeyChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}
QT_WARNING_POP
QT_END_MOC_NAMESPACE
